# -*- coding: utf-8 -*-
import law
import luigi
import logging
import os
import re
import shutil

from functools import partial
from itertools import product
from subprocess import PIPE

from law.util import colored

from analysis.tasks.Recombine import Recombine
from analysis.tasks.MergeFinal import MergeFinal

from analysis.framework import Task
from analysis.util import interruptable_popen, working_directory

logger = logging.getLogger(__name__)
working_directory = partial(working_directory, logger=logger)


class AddStatunc(Task, law.LocalWorkflow):
    merge_dir = luigi.Parameter(default='grids')
    steer_dir = luigi.Parameter()

    orders = luigi.ListParameter()
    observables = luigi.ListParameter()
    rivetid = luigi.Parameter()
    histoids = luigi.ListParameter()
    binsizefactors = luigi.ListParameter(default=['0'])
    multcoefffactors = luigi.ListParameter(default=['0'])

    def create_branch_map(self):
        """Branch data dicts. Each branch should correspond to exactly one output."""
        return {
            i : {'order': order, 'obs': obs, 'iobs': iobs}
            for i, (order, (iobs, obs)) in enumerate(product(
                self.orders, enumerate(self.observables)
            ))
        }

    def workflow_requires(self):
        return {'recombine': Recombine(), 'mergefinal': MergeFinal()}

    def requires(self):
        return Recombine(), MergeFinal()

    def output(self):
        out_dir = os.getenv("ANALYSIS_PATH") or os.getcwd()
        return law.LocalFileTarget(
            '{}/{}/{}/Combined/Final/{}.{}.{}_v25.tab.gz'.format(
                out_dir,
                self.merge_dir, self.name, self.process, self.branch_data['order'],
                self.branch_data['obs']
            )
        )

    def run(self):
        try:
            tabpath = '{}/{}/Combined/Final'.format(self.merge_dir, self.name)
            tabbase = '{}.{}.{}'.format(
                self.process, self.branch_data['order'], self.branch_data['obs']
            )
            tabfile = tabbase + '.tab.gz'
            outfile = tabbase + '_v25.tab.gz'
            strfile = tabbase + '.str'

            strtemplate = '{}/SteerModify.str'.format(self.steer_dir)
            strpathfile = os.path.join(tabpath, strfile)

            # Write per table strfile
            # Add RIVET_IDs if desired; requires SteerModify.str to contain a scenario description with RIVET_ID=
            if self.rivetid:
                logger.info('Found RIVET_ID: %s', self.rivetid)

            histoid = ""
            if self.histoids:
                if self.branch_data['iobs'] < len(self.histoids):
                    histoid = self.histoids[self.branch_data['iobs']]
                    logger.info('Found HISTO_ID: %s', histoid)
                else:
                    raise IndexError('Insufficient list size %s of histoids for iobs branch %s', len(self.histoids), self.branch_data['iobs'])

            binsizefactor = "1"
            if self.binsizefactors[0] != '0':
                if self.branch_data['iobs'] < len(self.binsizefactors):
                    binsizefactor = self.binsizefactors[self.branch_data['iobs']]
                    logger.info('Found BinSizeFactor: %s', binsizefactor)
                else:
                    raise IndexError('Insufficient list size %s of binsizefactors for iobs branch %s', len(self.binsizefactors), self.branch_data['iobs'])

            multcoefffactor = "1"
            if self.multcoefffactors[0] != '0':
                if self.branch_data['iobs'] < len(self.multcoefffactors):
                    multcoefffactor = self.multcoefffactors[self.branch_data['iobs']]
                    logger.info('Found MultcoeffFactor: %s', multcoefffactor)
                else:
                    raise IndexError('Insufficient list size %s of multcoefffactors for iobs branch %s', len(self.multcoefffactors), self.branch_data['iobs'])

            RID = 'RIVET_ID='
            with open(strpathfile, 'w') as ofile:
                with open(strtemplate, 'r') as ifile:
                    for line in ifile:
                        if RID in line and self.rivetid and histoid:
                            logger.debug(
                                'Changing input line with RIVET_ID:\n%s', line)
                            line = line.replace(RID, RID + self.rivetid + "/" + histoid)
                            logger.debug('Line now reads:\n%s', line)
                        ofile.write(line)

                logger.info(
                    'Appending modification specification to strfile: %s',
                    strpathfile
                )

                # Append information for grid-wise rescaling of bin widths and perturbative coefficients
                ofile.write('\n# Attaching grid-wise rescale factors for bin widths and perturbative coefficients\n')
                ofile.write('BinSizeFactor     {}        # Overall factor for bin sizes\n'.format(binsizefactor))
                ofile.write('MultCoeffFactor   {}        # Factor to scale stored coefficients for each observable bin!\n'.format(multcoefffactor))

                # Append information for adding statistical (numerical) uncertainty
                ofile.write('\n# Attaching missing InfoBlock information\n')
                ofile.write('InfoBlockFiles {\n')
                lofile = re.sub(self.branch_data['order'], 'LO', tabbase) + '.dat' + '\n'
                nlofile = re.sub(self.branch_data['order'], 'NLO_only', tabbase) + '.dat' + '\n'
                nnlofile = re.sub(self.branch_data['order'], 'NNLO_only', tabbase) + '.dat' + '\n'
                ofile.write('  ' + lofile)
                if self.branch_data['order'] != 'LO':
                    ofile.write('  ' + nlofile)
                if self.branch_data['order'] == 'NNLO':
                    ofile.write('  ' + nnlofile)
                ofile.write('}\n')
                ofile.write('InfoBlockOrders {\n')
                ofile.write('  LO\n')
                if self.branch_data['order'] != 'LO':
                    ofile.write('  NLO\n')
                if self.branch_data['order'] == 'NNLO':
                    ofile.write('  NNLO\n')
                ofile.write('}\n')

            # temporarily change to dir with final grids and dat files
            with working_directory(tabpath):
                command_line_args = [
                    'fnlo-tk-modify',
                    'steerfile={}'.format(strfile),
                    'InTable={}'.format(tabfile),
                    'OutTable={}'.format(outfile)
                ]

                logger.info("Running command: %s",
                    colored(' '.join(command_line_args), 'yellow'))
                code, out, error = interruptable_popen(command_line_args,
                                                       stdout=PIPE,
                                                       stderr=PIPE)
                logger.debug("Return code: %d", code)
                logger.debug("Standard output:\n%s\n", out)
                logger.debug("Standard error:\n%s\n", error)

            if (code == 0):
                logger.info("fnlo-tk-modify status: SUCCESS")
            else:
                raise RuntimeError("fnlo-tk-modify status: FAIL (exit code {})".format(code))

        # mark job as failed on error, make log entry and reraise
        except Exception as e:
            logger.error('%s raised during job execution: %s', e.__class__.__name__, e)
            raise
