# -*- coding: utf-8 -*-
import law
import luigi
import logging
import os
import re
import shutil

from itertools import product
from subprocess import PIPE

from law.util import colored

from analysis.tasks.Combine import Combine

from analysis.framework import Task, HTCondorWorkflow, SpaceSeparatedListParameter
from analysis.util import interruptable_popen

logger = logging.getLogger(__name__)


class MergeFastProd(Task, HTCondorWorkflow, law.LocalWorkflow):

    # DS: is this supposed to be run remotely at all?
    # the implementation appears to read files that are only
    # available locally?
    # KR: This is a very time consuming step, where each observable grid for one channel has to be read
    #     consecutively, e.g. several thousands for RRa. Therefore submission to HTcondor unpacking the tar.gz
    #     on the grid storage was foreseen, but not implemented. In the meantime it seems more interesting to me
    #     to eliminate this bottleneck by improving the fastNLO merge procedure such that "consecutive" can be
    #     transformed into a "paralellised pre-merging" and one final merge step. TBD.

    merge_dir = luigi.Parameter(default='grids')

    default_channels = ["LO", "R", "V", "RRa", "RRb", "RV", "VV"]
    channels = SpaceSeparatedListParameter(
        default=default_channels,
        cls=luigi.Parameter,
        choices=("LO", "R", "V", "RR", "RRa", "RRb", "RV", "VV")
    )
    weightorder = luigi.Parameter()
    observables = luigi.ListParameter()

    def create_branch_map(self):
        """Branch data dicts. Each branch should correspond to exactly one output."""
        return {
            i : {'channel': channel, 'obs': obs}
            for i, (channel, obs) in enumerate(product(
                self.channels,
                self.observables
            ))
        }

    def workflow_requires(self):
        return {'combine': Combine()}

    def requires(self):
        return Combine()

    def output(self):
        out_dir = os.getenv("ANALYSIS_PATH") or os.getcwd()
        return law.LocalFileTarget(
            '{}/{}/{}/Combined/Final/{}.{}.{}.tab.gz'.format(
                out_dir,
                self.merge_dir, self.name, self.process, self.branch_data['channel'],
                self.branch_data['obs']
            )
        )

    def run(self):
        # ensure output parent directory exists
        logger.debug("Creating output parent directory: %s", self.output().parent.path)
        self.output().parent.touch()

        outfile = self.output().path
        logfile = re.sub('.tab.gz$', '.merge2.log', outfile)
        errfile = re.sub('.tab.gz$', '.merge2.err', outfile)

        weightfile = '{}/{}/Combined/Final/{}.{}.APPLfast.txt'.format(
            self.merge_dir, self.name, self.weightorder, self.branch_data['obs']
        )

        # DS: why are the (Copy)FastProd tasks that produce these tables not added
        # as dependencies?
        # KR: In addition to the APPLfast weight files produced by the Combine task,
        #     the MergeFastProd task requires all grids produced for all channels and observables
        #     similarly as the Combine task requires all dat files for all channels and observables.
        #     For efficiency reasons therefore only the successful Combine step is required.
        #     ATTENTION: In case of good dat files used in Combine but unreadable corresponding tab.gz
        #                grids, MergeFastProd fails as it should to avoid discrepancies between
        #                combined NNLOJET and merged fastNLO results.
        #                Therefore it is STRONGLY RECOMMENDED to ensure the readability of all grids
        #                by completing the FnloCppread task BEFORE even running the Combine task and
        #                by removing all jobs and outputs containing a malformed result.
        #                Fortunately this happens rarely, but it is quite annoying when it occurs.
        tableglob = '{}/{}/{}/*.{}.s*.tab.gz'.format(
            self.merge_dir, self.name, self.branch_data['channel'], self.branch_data['obs']
        )

        command_line = 'fnlo-tk-merge2 -w NNLOJET {} {} {}'.format(weightfile, tableglob, outfile)

        logger.info("Running command: %s", colored(command_line, 'yellow'))
        code, out, error = interruptable_popen(command_line, shell=True, stdout=PIPE, stderr=PIPE)

        logger.debug("Return code: %d", code)
        logger.debug("Standard output:\n%s\n", out)
        logger.debug("Standard error:\n%s\n", error)

        logger.debug("Dumping standard out to file:  %s", logfile)
        with open(logfile, 'w') as f:
            f.write(out)

        if error:
            logger.debug("Dumping standard error to file:  %s", errfile)
            with open(errfile, 'w') as f:
                f.write(error)

        if (code == 0):
            logger.info("fnlo-tk-merge2 status: SUCCESS")
        else:
            error_message = "fnlo-tk-merge2 status: FAIL (exit code {})".format(code)
            logger.error(error_message)
            raise RuntimeError(error_message)
